/*
 * Asterisk -- An open source telephony toolkit.
 *
 * Copyright (C) 1999 - 2005, Digium, Inc.
 *
 * Mark Spencer <markster@digium.com>
 *
 * Includes code and algorithms from the Zapata library.
 *
 * See http://www.asterisk.org for more information about
 * the Asterisk project. Please do not directly contact
 * any of the maintainers of this project for assistance;
 * the project provides a web site, mailing lists and IRC
 * channels for your use.
 *
 * This program is free software, distributed under the terms of
 * the GNU General Public License Version 2. See the LICENSE file
 * at the top of the source tree.
 */

/*!
 * \file
 * \brief Curl based asterisk CDR.
 *
 * \author Alex Baron <alex.baron@rgroup.com.au>
 *
 * \ingroup cdr_drivers
 */

#include "asterisk.h"
ASTERISK_FILE_VERSION(__FILE__, "$Revision: 0 $")

#include <curl/curl.h>

#include "asterisk/config.h"
#include "asterisk/cdr.h"
#include "asterisk/module.h"
#include "asterisk/utils.h"
#include "asterisk/lock.h"
#include "asterisk/threadstorage.h"
#include "asterisk/logger.h"
#include "asterisk/threadstorage.h"
#include "asterisk/pbx.h"

AST_THREADSTORAGE(query_buf);
AST_THREADSTORAGE(result_buf);


#define DATE_FORMAT "%Y-%m-%d %T"

static char *name = "cdr_curl";
static char endpointurl[1024] = "";
static int loaded = 0;
static const char config[] = "cdr_curl.conf";


/* Append a string to an existing char array */
static int append_string(char *buf, const char *s, size_t bufsize)
{
	int pos = strlen(buf), spos = 0, error = -1;

	if (pos >= bufsize - 4)
		return -1;

	while(pos < bufsize - 3) {
		if (!s[spos]) {
			error = 0;
			break;
		}
		buf[pos++] = s[spos];
		spos++;
	}

	buf[pos++] = '\0';

	return error;
}

/* Append a date to an existing char array */
static int append_date_and_encode(char *buf, struct timeval when, size_t bufsize)
{
	char tmp[80] = "";
	struct ast_tm tm;

	if (strlen(buf) > bufsize - 3)
		return -1;

	if (ast_tvzero(when)) {
		strncat(buf, ",", bufsize - strlen(buf) - 1);
		return 0;
	}

	ast_localtime(&when, &tm, "GMT");
	ast_strftime(tmp, sizeof(tmp), DATE_FORMAT, &tm);

	char etmp[1024];
	ast_uri_encode(tmp, etmp, sizeof(etmp), ast_uri_http);
	
	return append_string(buf, etmp, bufsize);
}

/* Append a int to an existing char array */
static int append_int(char *buf, int s, size_t bufsize)
{
	char tmp[32];
	int pos = strlen(buf);

	snprintf(tmp, sizeof(tmp), "%d", s);

	if (pos + strlen(tmp) > bufsize - 3)
		return -1;

	strncat(buf, tmp, bufsize - strlen(buf) - 1);
	pos = strlen(buf);
	buf[pos++] = '&';
	buf[pos++] = '\0';

	return 0;
}

/* Helper to format the values into a key=value& format. */
static int append_string_query_param(const char *key, const char *value, char *buf, size_t bufsize)
{
	char tmp[1024] = "";
	ast_uri_encode(value, tmp, sizeof(tmp), ast_uri_http);
	
	append_string(buf, key, bufsize);
	append_string(buf, "=", bufsize);
	append_string(buf, tmp, bufsize);
	return append_string(buf, "&", bufsize);
}

/* Helper to format the values into a key=value& format. */
static int append_datetime_query_param(const char *key, struct timeval value, char *buf, size_t bufsize)
{
	append_string(buf, key, bufsize);
	append_string(buf, "=", bufsize);
	append_date_and_encode(buf, value, bufsize);
	return append_string(buf, "&", bufsize);
}

/* Helper to format the values into a key=value& format. */
static int append_int_query_param(const char *key, int value, char *buf, size_t bufsize)
{
	append_string(buf, key, bufsize);
	append_string(buf, "=", bufsize);
	append_int(buf, value, bufsize);
	return append_string(buf, "&", bufsize);
}


/* load the configuration */
static int load_config(int reload)
{
	struct ast_config *cfg;
	const char *tmp;
	struct ast_flags config_flags = { reload ? CONFIG_FLAG_FILEUNCHANGED : 0 };

	if (!(cfg = ast_config_load(config, config_flags)) || cfg == CONFIG_STATUS_FILEINVALID) {
		ast_log(LOG_WARNING, "unable to load config: %s\n", config);
		return 0;
	} else if (cfg == CONFIG_STATUS_FILEUNCHANGED)
		return 1;

	if ((tmp = ast_variable_retrieve(cfg, "curl", "url"))) 
	{
		ast_copy_string(endpointurl, tmp, sizeof(endpointurl));
	}
	else
	{
		return AST_MODULE_LOAD_DECLINE; /* we are missing the url to send data to, this is a show stopper */
	}

	ast_log(LOG_NOTICE, "Found root url in config: '%s'.\n", endpointurl);

	ast_config_destroy(cfg);

	return 1;
}

/* occours each time cdr is written */
static int cdr_curl(struct ast_cdr *cdr)
{
	if (!ast_custom_function_find("CURL")) {
		ast_log(LOG_ERROR, "func_curl.so must be loaded in order to use res_config_curl.so!!\n");
		return -1;
	}

	char buf[2048];
	struct ast_str *query, *buffer;
	
	/* Not sure whats going on here, its from the res_config_curl, there are statements at the top of the file that relate. */
	if (!(query = ast_str_thread_get(&query_buf, 16))) { return -1; }
	if (!(buffer = ast_str_thread_get(&result_buf, 16))) { return -1; }

	/* Tell the user where we are sending the response. */
	ast_log(LOG_NOTICE, "Url to send data to is '%s'.\n", endpointurl);

	buf[0] = '\0';
	
	/* Set the parameters as here http://www.asterisk.org/doxygen/asterisk1.0/structast__cdr.html#8098d02138e27b57d184b260b6669eed  */
	append_string_query_param("clid", cdr->clid, buf, sizeof(buf));
	append_string_query_param("src", cdr->src, buf, sizeof(buf));
	append_string_query_param("dst", cdr->dst, buf, sizeof(buf));
	append_string_query_param("dcontext", cdr->dcontext, buf, sizeof(buf));
	append_string_query_param("channel", cdr->channel, buf, sizeof(buf));
	append_string_query_param("dstchannel", cdr->dstchannel, buf, sizeof(buf));
	append_string_query_param("lastapp", cdr->lastapp, buf, sizeof(buf));
	append_string_query_param("lastdata", cdr->lastdata, buf, sizeof(buf));
	append_datetime_query_param("start", cdr->start, buf, sizeof(buf));
	append_datetime_query_param("answer", cdr->answer, buf, sizeof(buf));
	append_datetime_query_param("end", cdr->end, buf, sizeof(buf));
	append_int_query_param("duration", cdr->duration, buf, sizeof(buf));
	append_int_query_param("billsec", cdr->billsec, buf, sizeof(buf));
	append_string_query_param("disposition", ast_cdr_disp2str(cdr->disposition), buf, sizeof(buf));
	append_string_query_param("amaflags", ast_cdr_flags2str(cdr->amaflags), buf, sizeof(buf));
	append_string_query_param("accountcode", cdr->accountcode, buf, sizeof(buf));
	append_string_query_param("flags", ast_cdr_flags2str(cdr->flags), buf, sizeof(buf));
	append_string_query_param("uniqueid", cdr->uniqueid, buf, sizeof(buf));
	append_string_query_param("userfield", cdr->userfield, buf, sizeof(buf));
	
	/* Build the actual http get query. */
	ast_str_set(&query, 0, "${CURL(%s?%s)}", endpointurl, buf);

	/* Print Out Query */
	ast_log(LOG_NOTICE, "Curl CDR query is: '%s'\n", ast_str_buffer(query));

	/* Call Curl */
	ast_str_substitute_variables(&buffer, 0, NULL, ast_str_buffer(query));

	/* Print Out Response */
    ast_log(LOG_NOTICE, "Curl CDR response is: '%s'\n", ast_str_buffer(buffer));

    return 0;
}


/* Occours when asterisk loads the module */
static int load_module(void)
{
	int res;

	if(!load_config(0)) 
		return AST_MODULE_LOAD_DECLINE;

	if ((res = ast_cdr_register(name, "Curl CDR Handler", cdr_curl))) {
		ast_log(LOG_ERROR, "Unable to register CDR CURL handling\n");
	} else {
		loaded = 1;
	}

	return res;
}

/* Occours when asterisk unloads the module. */
static int unload_module(void)
{
	/* Unregister the curl method */
	ast_cdr_unregister(name);

	loaded = 0;

	return 0;
}

/* occours when asterisk reloads the module. */ 
static int reload(void)
{	
	if(load_config(1)) 
	{
		loaded = 1;
	}
	else
	{
		loaded = 0;
		ast_log(LOG_WARNING, "No [curl] section in cdr_curl.conf.  Unregistering backend.\n");
		ast_cdr_unregister(name);	
	}

	return 0;
}

AST_MODULE_INFO(ASTERISK_GPL_KEY, AST_MODFLAG_DEFAULT, "Curl CDR Backend",	
	.load = load_module,	
	.unload = unload_module,	
	.reload = reload,		
);
